@@ -1,6 +1,6 @@
-/* prototype_7.c    2025-07-23 */
+/* prototype_7.c    2026-01-26 */
 
-/* Copyright 2024-2025 Emmanuel Paradis */
+/* Copyright 2024-2026 Emmanuel Paradis */
 
 /* This file is part of the R-package `hann'. */
 /* See the file ../DESCRIPTION for licensing issues. */
@@ -15,6 +15,9 @@
 static int N, K, C, np1, npar, KC, mc_cores;
 static double target, BETA;
 
+/* arrays used at each iteration */
+double *O7, *Deviation, *bar_O7;
+
 /*
   control_list[0]: quasiNewton
   control_list[1]: fullHessian
@@ -52,9 +55,7 @@ double objfun_7_OMP(double *PARA, double *sigma_xi, int *E, double *GRAD, int ev
     int mu, i, j, k, idx;
     double val = 0, s, tmp;
 
-    double *w, *bias, *grad_W, *grad_bias, *O, *Deviation, *bar_O;
-
-    O = (double*)R_alloc(KC, sizeof(double));
+    double *w, *bias, *grad_W, *grad_bias;
 
     w = PARA;
     bias = PARA + np1;
@@ -62,28 +63,25 @@ double objfun_7_OMP(double *PARA, double *sigma_xi, int *E, double *GRAD, int ev
     if (eval_grad) {
 	grad_W = GRAD;
 	grad_bias = GRAD + np1;
-
-	Deviation = (double*)R_alloc(KC, sizeof(double));
-	bar_O = (double*)R_alloc(KC, sizeof(double));
     }
 
     /* 2. integrate the signals from the N input neurons to the C
        classification neurons */
-    fast_mat_prod_0(sigma_xi, w, O, K, C, N);
+    fast_mat_prod_0(sigma_xi, w, O7, K, C, N);
 
 #pragma omp parallel for// shared(KC, O, bias, BETA) private(i)
     for (i = 0; i < KC; i++)
-	O[i] = tanh(BETA * O[i] + bias[i/K]); // integer division (i/K)
+	O7[i] = my_tanh(BETA * O7[i] + bias[i/K]); // integer division (i/K)
 
     if (control_list[2]) {
-	k = do_error_rate(E, O, K, C);
-	Rprintf("Error rate = %d / %d\n", k, K);
+	k = do_error_rate(E, O7, K, C);
+	Rprintf("\tError rate = %d / %d", k, K);
     }
 
     if (eval_grad) {
 #pragma omp parallel for// shared(KC, O, bar_O) private(i)
 	for (i = 0; i < KC; i++)
-	    bar_O[i] = bar(O[i]);
+	    bar_O7[i] = bar(O7[i]);
     }
 
 /*
@@ -112,12 +110,12 @@ double objfun_7_OMP(double *PARA, double *sigma_xi, int *E, double *GRAD, int ev
 
     if (eval_grad) {
 	for (i = 0; i < KC; i++) {
-	    Deviation[i] = s = O[i] - E[i];
+	    Deviation[i] = s = O7[i] - E[i];
 	    val += s * s;
 	}
     } else {
 	for (i = 0; i < KC; i++)
-	    val += pow(O[i] - E[i], 2);
+	    val += pow(O7[i] - E[i], 2);
 	return val;
     }
 
@@ -136,7 +134,7 @@ double objfun_7_OMP(double *PARA, double *sigma_xi, int *E, double *GRAD, int ev
 	    idx = j * K;
 	    for (mu = 0; mu < K; mu++) {
 		// idx = mu + j*K;
-		s += sigma_xi[mu + i*K] * bar_O[idx] * Deviation[idx];
+		s += sigma_xi[mu + i*K] * bar_O7[idx] * Deviation[idx];
 		idx++;
 	    }
 	    grad_W[k++] = tmp * s; // k = i + j*N
@@ -148,7 +146,7 @@ double objfun_7_OMP(double *PARA, double *sigma_xi, int *E, double *GRAD, int ev
 	s = 0;
 	for (mu = 0; mu < K; mu++) {
 	    idx = mu + j*K;
-	    s += bar_O[idx] * Deviation[idx];
+	    s += bar_O7[idx] * Deviation[idx];
 	}
 	grad_bias[j]  = 2 * s;
     }
@@ -189,9 +187,7 @@ double objfun_7(double *PARA, double *sigma_xi, int *E, double *GRAD, int eval_g
     int mu, i, j, k, idx;
     double val = 0, s, tmp;
 
-    double *w, *bias, *grad_W, *grad_bias, *O, *Deviation, *bar_O;
-
-    O = (double*)R_alloc(KC, sizeof(double));
+    double *w, *bias, *grad_W, *grad_bias;
 
     w = PARA;
     bias = PARA + np1;
@@ -199,36 +195,33 @@ double objfun_7(double *PARA, double *sigma_xi, int *E, double *GRAD, int eval_g
     if (eval_grad) {
 	grad_W = GRAD;
 	grad_bias = GRAD + np1;
-
-	Deviation = (double*)R_alloc(KC, sizeof(double));
-	bar_O = (double*)R_alloc(KC, sizeof(double));
     }
 
     /* 2. integrate the signals from the N input neurons to the C
        classification neurons */
-    fast_mat_prod_0(sigma_xi, w, O, K, C, N);
+    fast_mat_prod_0(sigma_xi, w, O7, K, C, N);
 
     for (i = 0; i < KC; i++)
-	O[i] = tanh(BETA * O[i] + bias[i/K]); // integer division (i/K)
+	O7[i] = my_tanh(BETA * O7[i] + bias[i/K]); // integer division (i/K)
 
     if (control_list[2]) {
-	k = do_error_rate(E, O, K, C);
-	Rprintf("Error rate = %d / %d\n", k, K);
+	k = do_error_rate(E, O7, K, C);
+	Rprintf("\tError rate = %d / %d", k, K);
     }
 
     if (eval_grad) {
 	for (i = 0; i < KC; i++)
-	    bar_O[i] = bar(O[i]);
+	    bar_O7[i] = bar(O7[i]);
     }
 
     if (eval_grad) {
 	for (i = 0; i < KC; i++) {
-	    Deviation[i] = s = O[i] - E[i];
+	    Deviation[i] = s = O7[i] - E[i];
 	    val += s * s;
 	}
     } else {
 	for (i = 0; i < KC; i++)
-	    val += pow(O[i] - E[i], 2);
+	    val += pow(O7[i] - E[i], 2);
 	return val;
     }
 
@@ -247,7 +240,7 @@ double objfun_7(double *PARA, double *sigma_xi, int *E, double *GRAD, int eval_g
 	    idx = j * K;
 	    for (mu = 0; mu < K; mu++) {
 		// idx = mu + j*K;
-		s += sigma_xi[mu + i*K] * bar_O[idx] * Deviation[idx];
+		s += sigma_xi[mu + i*K] * bar_O7[idx] * Deviation[idx];
 		idx++;
 	    }
 	    grad_W[k++] = tmp * s; // k = i + j*N
@@ -259,7 +252,7 @@ double objfun_7(double *PARA, double *sigma_xi, int *E, double *GRAD, int eval_g
 	s = 0;
 	for (mu = 0; mu < K; mu++) {
 	    idx = mu + j*K;
-	    s += bar_O[idx] * Deviation[idx];
+	    s += bar_O7[idx] * Deviation[idx];
 	}
 	grad_bias[j]  = 2 * s;
     }
@@ -285,7 +278,7 @@ void do_Hessian(double *PARA, double *sigma_xi, int *E,
 {
     int i, j, k = 0;
     double delta = 1E-8, *new_grad, tmp;
-    new_grad = (double*)R_alloc(npar, sizeof(double));
+    new_grad = malloc(npar * sizeof(double));
 
     for (i = 0; i < npar; i++) {
 	tmp = PARA[i];
@@ -295,6 +288,8 @@ void do_Hessian(double *PARA, double *sigma_xi, int *E,
 	    hessian[k++] = (new_grad[j] - GRAD[j]) / delta;
 	PARA[i] = tmp;
     }
+
+    free(new_grad);
 }
 
 #define COMPUTE_PROPOSED_PARA				\
@@ -340,8 +335,7 @@ double optimize_7(double *PARA, int *sigma, int *xi,
     ptr_grad[1] = new_GRAD;
 
     /* 1. calculate the intermediate products 'sigma_i * xi_{mu,i}'
-       only once (may take a lot of memory if K is large...
-       => can overwrite 'xi' instead?) */
+       only once */
     k = 0;
     for (i = 0; i < N; i++) {
 	s = (double)sigma[i];
@@ -353,10 +347,10 @@ double optimize_7(double *PARA, int *sigma, int *xi,
     }
 
     /* do a first iteration to get the gradients: */
+    if (!quiet) Rprintf("\riteration 0");
     res = FUN(PARA, sigma_xi, expec, GRAD, 1);
-
-    if (!quiet) Rprintf("INITIALIZATION -- iteration %d\tobj_fun = %f\n", iter, res);
-    if (!quiet) Rprintf("Gradients done.\n");
+    if (!quiet) Rprintf("\tobj_fun = %.3f", res);
+    if (control_list[2]) Rprintf("\n");
 
     for (;;) {
 	if (iter >= iterlim) break;
@@ -437,18 +431,23 @@ update:
 	    memset(ptr_para[Switch] + np1, 0,
 		   C * sizeof(double));
 	}
+
+	if (!quiet) Rprintf("\riteration %d", iter + 1);
+
 	/* update the gradients (and 'res') */
 	res = FUN(ptr_para[Switch], sigma_xi, expec,
 		  ptr_grad[Switch], 1);
+
+	if (!quiet) Rprintf("\tobj_fun = %.3f", res);
+	if (control_list[2]) Rprintf("\n");
+
 	if (res < target) break;
 	R_CheckUserInterrupt();
 	iter++;
     }
 
     if (!quiet && iterlim) Rprintf("\n");
-
     if (Switch) memcpy(PARA, new_PARA, npar * sizeof(double));
-
     return res;
 }
 
@@ -496,6 +495,11 @@ SEXP test_7(SEXP W, SEXP BIAS, SEXP SIGMA, SEXP XI, SEXP EXPEC,
     memcpy(PARA, w, np1 * sizeof(double));
     memcpy(PARA + np1, bias, C * sizeof(double));
 
+    /* global arrays */
+    O7 = (double*)R_alloc(KC, sizeof(double));
+    Deviation = (double*)R_alloc(KC, sizeof(double));
+    bar_O7 = (double*)R_alloc(KC, sizeof(double));
+
     val = optimize_7(PARA, sigma, xi, expec, INTEGER(ITERLIM)[0],
 		     INTEGER(QUIET)[0]);
 
